#!/usr/bin/env python
# django_shell.py

from django.core.management import call_command
from boot_django import boot_django, APP_NAME  # noqa


boot_django()
print(f'running canary tests for {APP_NAME}')
call_command('test', APP_NAME, '--tag=canary', verbosity=2)

# TODO: consider collecting switches and arguments
#  from the command line (like -v or a specific test module
#  or function) and passing them on to call command



from datetime import datetime
from faker import Faker

fake = Faker()


def create_test_alert(topic=None):
    alert = {
        'id': fake.pyint(min_value=2458000, max_value=2463000),
        'alert_identifier': fake.pystr(),
        'alert_timestamp': datetime.strftime(fake.date_time_this_month(), '%Y-%m-%d %H:%M:%S'),
        'topic': fake.pyfloat(min_value=15, max_value=23),
        'right_ascension': fake.pyfloat(min_value=0, max_value=360),
        'declination': fake.pyfloat(min_value=0, max_value=360),
        'extracted_fields': {},  # fake.pydict(nb_elements=0),
        'message': {},  # fake.pydict(nb_elements=0)
    }

    if topic:
        alert['topic'] = topic

    return alert



import requests

from confluent_kafka import KafkaException
from crispy_forms.layout import Column, Div, Fieldset, HTML, Layout, Row
from django import forms
from django.conf import settings
from hop import Stream
from hop.auth import Auth

from tom_alerts.alerts import GenericAlert, GenericBroker, GenericQueryForm, GenericUpstreamSubmissionForm
from tom_alerts.exceptions import AlertSubmissionException
from tom_targets.models import Target

SCIMMA_URL = 'http://skip.dev.hop.scimma.org'
SCIMMA_API_URL = f'{SCIMMA_URL}/api'


class SCIMMAQueryForm(GenericQueryForm):
    keyword = forms.CharField(required=False, label='Keyword search')
    topic = forms.MultipleChoiceField(choices=[], required=False, label='Topic')
    cone_search = forms.CharField(required=False, label='Cone Search', help_text='RA, Dec, radius in degrees')
    polygon_search = forms.CharField(required=False, label='Polygon Search',
                                     help_text='Comma-separated pairs of space-delimited coordinates (degrees)')
    alert_timestamp_after = forms.DateTimeField(required=False, label='Datetime lower')
    alert_timestamp_before = forms.DateTimeField(required=False, label='Datetime upper')
    event_trigger_number = forms.CharField(required=False, label='LVC Trigger Number')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['topic'].choices = self.get_topic_choices()

        self.helper.layout = Layout(
            self.common_layout,
            Fieldset(
                '',
                Div('topic'),
                Div('keyword')
            ),
            Fieldset(
                'Time Filters',
                Row(
                    Column('alert_timestamp_after'),
                    Column('alert_timestamp_before')
                )
            ),
            Fieldset(
                'Spatial Filters',
                Div('cone_search'),
                Div('polygon_search')
            ),
            HTML('<hr>'),
            Fieldset(
                'LVC Trigger Number',
                HTML('''
                    <p>
                    The LVC Trigger Number filter will only search the LVC topic. Please be aware that any topic
                    selections will be ignored.
                    </p>
                '''),
                'event_trigger_number'
            )
        )

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get('event_trigger_number') and cleaned_data.get('topic'):
            raise forms.ValidationError('Topic filter cannot be used with LVC Trigger Number filter.')
        return cleaned_data

    @staticmethod
    def get_topic_choices():
        response = requests.get(f'{SCIMMA_API_URL}/topics')
        response.raise_for_status()
        return [(result['id'], result['name']) for result in response.json()['results']]


class SCIMMAUpstreamSubmissionForm(GenericUpstreamSubmissionForm):
    topic = forms.CharField(required=False, max_length=100, widget=forms.HiddenInput())


class SCIMMABroker(GenericBroker):
    """
    This is a prototype interface to the skip db built by SCIMMA
    """

    name = 'SCIMMA'
    form = SCIMMAQueryForm
    alert_submission_form = SCIMMAUpstreamSubmissionForm

    def _request_alerts(self, parameters):
        response = requests.get(f'{SCIMMA_API_URL}/alerts/',
                                params={**parameters},
                                headers=settings.BROKERS['SCIMMA'])
        response.raise_for_status()
        return response.json()

    def fetch_alerts(self, parameters):
        parameters['page_size'] = 20
        result = self._request_alerts(parameters)
        return iter(result['results'])

    def fetch_alert(self, alert_id):
        url = f'{SCIMMA_API_URL}/alerts/{alert_id}'
        response = requests.get(url, headers=settings.BROKERS['SCIMMA'])
        response.raise_for_status()
        parsed = response.json()
        return parsed

    def to_generic_alert(self, alert):
        score = alert['message'].get('rank', 0) if alert['topic'] == 'lvc-counterpart' else ''
        return GenericAlert(
            url=f'{SCIMMA_API_URL}/alerts/{alert["id"]}',
            id=alert['id'],
            # This should be the object name if it is in the comments
            name=alert['alert_identifier'],
            ra=alert['right_ascension'],
            dec=alert['declination'],
            timestamp=alert['alert_timestamp'],
            # Well mag is not well defined for XRT sources...
            mag=0.0,
            score=score  # Not exactly what score means, but ish
        )

    def to_target(self, alert):
        # Galactic Coordinates come in the format:
        # "gal_coords": "76.19,  5.74 [deg] galactic lon,lat of the counterpart",
        gal_coords = [None, None]
        if alert['topic'] == 'lvc-counterpart':
            gal_coords = alert['message'].get('gal_coords', '').split('[')[0].split(',')
            gal_coords = [float(coord.strip()) for coord in gal_coords]
        return Target.objects.create(
            name=alert['alert_identifier'],
            type='SIDEREAL',
            ra=alert['right_ascension'],
            dec=alert['declination'],
            galactic_lng=gal_coords[0],
            galactic_lat=gal_coords[1],
        )

    def submit_upstream_alert(self, target=None, observation_record=None, **kwargs):
        """
        Submits target and observation record as Hopskotch alerts.

        :param target: ``Target`` object to be converted to an alert and submitted upstream
        :type target: ``Target``

        :param observation_record: ``ObservationRecord`` object to be converted to an alert and submitted upstream
        :type observation_record: ``ObservationRecord``

        :param \\**kwargs:
            See below

        :Keyword Arguments:
            * *topic* (``str``): Hopskotch topic to submit the alert to.

        :returns: True or False depending on success of message submission
        :rtype: bool

        :raises:
            AlertSubmissionException: If topic is not provided to the function and a default is not provided in
                                      settings
        """
        creds = settings.BROKERS['SCIMMA']
        stream = Stream(auth=Auth(creds['hopskotch_username'], creds['hopskotch_password']))
        stream_url = creds['hopskotch_url']
        topic = kwargs.get('topic') if kwargs.get('topic') else creds['default_hopskotch_topic']

        if not topic:
            raise AlertSubmissionException(f'Topic must be provided to submit alert to {self.name}')

        try:
            with stream.open(f'kafka://{stream_url}:9092/{topic}', 'w') as s:
                if target:
                    message = {'type': 'target', 'target_name': target.name, 'ra': target.ra, 'dec': target.dec}
                    s.write(message)
                if observation_record:
                    message = {'type': 'observation', 'status': observation_record.status,
                               'parameters': observation_record.parameters,
                               'target_name': observation_record.target.name,
                               'ra': observation_record.target.ra, 'dec': observation_record.target.dec,
                               'facility': observation_record.facility}
                    s.write(message)
        except KafkaException as e:
            raise AlertSubmissionException(f'Submission to Hopskotch failed: {e}')

        return True



from django import template

from tom_scimma.scimma import SCIMMAUpstreamSubmissionForm


register = template.Library()


@register.inclusion_tag('tom_scimma/partials/submit_upstream_scimma_form.html')
def submit_upstream_scimma_form(target=None, observation_record=None, redirect_url=None, topic=None):
    """
    Renders a button to submit an alert upstream to a broker. At least one of target/obs record should be given.

    :param broker: The name of the broker to which the button will lead, as in the name field of the broker module.
    :type broker: str

    :param target: The target to be submitted as an alert, if any.
    :type target: ``Target``

    :param observation_record: The observation record to be submitted as an alert, if any.
    :type observation_record: ``ObservationRecord``

    :param topic: The topic to submit the alerts to.
    :type topic: str

    :param redirect_url:
    :type redirect_url: str
    """

    form = SCIMMAUpstreamSubmissionForm(broker='SCIMMA', initial={
        'target': target,
        'observation_record': observation_record,
        'redirect_url': redirect_url,
        'topic': topic
    })

    return {
        'submit_upstream_form': form
    }



from django.test import tag, TestCase

from tom_scimma.scimma import SCIMMABroker
from tom_targets.models import Target


@tag('canary')
class TestSCIMMAModuleCanary(TestCase):
    """NOTE: To run these tests in your venv: python ./tom_scimma/tests/run_tests.py"""

    def setUp(self):
        self.broker = SCIMMABroker()
        self.expected_keys = ['id', 'alert_identifier', 'alert_timestamp', 'topic', 'right_ascension', 'declination',
                              'right_ascension_sexagesimal', 'declination_sexagesimal', 'role', 'extracted_fields',
                              'message', 'created', 'modified']

    def test_boilerplate(self):
        self.assertTrue(True)

    def test_fetch_alerts(self):
        """Test fetch_alerts"""
        response = self.broker.fetch_alerts({'topic': 1})
        alerts = []
        for alert in response:
            alerts.append(alert)
            for key in self.expected_keys:
                self.assertTrue(key in alert.keys())
        self.assertEqual(len(alerts), 20)

    def test_fetch_alert(self):
        """Test fetch_alert"""
        alert = self.broker.fetch_alert(6911)
        self.assertEqual(alert['right_ascension'], 242.322)
        for key in self.expected_keys:
            self.assertTrue(key in alert.keys())

    def test_submit_upstream_alert(self):
        """Test submit_upstream_alert"""
        t = Target.objects.create(name='canary test target', ra=1, dec=2)
        response = self.broker.submit_upstream_alert(target=t, observation_record=None, topic='TOMToolkit.test')

        self.assertTrue(response)



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='tom-scimma',
    description='Skip/Hopskotch broker module for the TOM Toolkit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://scimma.org',
    author='TOM Toolkit Project',
    author_email='dcollom@lco.global',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    keywords=['tomtoolkit', 'astronomy', 'astrophysics', 'cosmology', 'science', 'fits', 'observatory', 'scimma'],
    packages=find_packages(),
    use_scm_version=True,  # use_scm_version and setup_requires setuptools_scm are required for automated releases
    setup_requires=['setuptools_scm', 'wheel'],
    install_requires=[
        'tomtoolkit~=2.6',
        'hop-client~=0.3'
    ],
    extras_require={
        'test': ['factory_boy~=3.1']
    },
    include_package_data=True,
)



import json
from requests import Response
from unittest.mock import mock_open, patch

from confluent_kafka import KafkaException
from django.core.exceptions import NON_FIELD_ERRORS
from django.test import override_settings, TestCase

from tom_alerts.exceptions import AlertSubmissionException
from tom_scimma.scimma import SCIMMAQueryForm, SCIMMABroker
from tom_scimma.tests.utils import create_test_alert
from tom_targets.models import Target


class TestSCIMMAQueryForm(TestCase):
    """
    NOTE: to run these tests in your venv: python ./tom_scimma/tests/run_tests.py
    """
    def setUp(self):
        mock_topic_choices = {
            'count': 1,
            'results': [
                {'id': 1, 'name': 'gcn'},
                {'id': 2, 'name': 'lvc-counterpart'}
            ]
        }
        self.mock_response = Response()
        self.mock_response._content = str.encode(json.dumps(mock_topic_choices))
        self.mock_response.status_code = 200

    def test_boilerplate(self):
        """Make sure the testing infrastructure is working."""
        self.assertTrue(True)

    @patch('tom_scimma.scimma.requests.get')
    def test_clean_topic_and_trigger_number(self, mock_requests_get):
        """Test that an error is thrown when both Topic and LVC Trigger Number are included in form submission."""
        mock_requests_get.return_value = self.mock_response

        form = SCIMMAQueryForm({'query_name': 'Test SCIMMA Query',
                                'broker': 'SCIMMA',
                                'topic': [1],
                                'event_trigger_number': 'S190426'})
        form.is_valid()
        self.assertTrue(form.has_error(NON_FIELD_ERRORS))
        self.assertIn('Topic filter cannot be used with LVC Trigger Number filter.', form.non_field_errors())


@override_settings(BROKERS={'SCIMMA': {'api_key': '', 'hopskotch_username': '', 'hopskotch_password': '',
                                       'hopskotch_url': 'test_url', 'default_hopskotch_topic': 'default'}
                            })
class TestSCIMMABrokerClass(TestCase):
    """NOTE: To run these tests in your venv: python ./tom_scimma/tests/run_tests.py"""

    def setUp(self):
        # Create 20 alerts
        self.alerts = [create_test_alert() for i in range(0, 20)]

    @patch('tom_scimma.scimma.requests.get')
    def test_fetch_alerts(self, mock_requests_get):
        """Test the SCIMMA-specific fetch_alerts logic."""
        mock_response = Response()
        mock_response._content = str.encode(json.dumps({'results': self.alerts}))
        mock_response.status_code = 200
        mock_requests_get.return_value = mock_response

        alerts_response = SCIMMABroker().fetch_alerts({})

        alerts = [alert for alert in alerts_response]
        self.assertEqual(len(alerts), 20)
        self.assertEqual(alerts[0], self.alerts[0])

    def test_to_generic_alert_any_topic(self):
        """Test the SCIMMA-specific to_generic_alert logic."""
        test_alert = self.alerts[0]
        test_alert['topic'] = 'gcn'
        test_alert['message'] = {'rank': '3'}
        generic_alert = SCIMMABroker().to_generic_alert(test_alert)

        # NOTE: The string is hardcoded as a sanity check to ensure that the string is reviewed if it changes
        self.assertEqual(generic_alert.url, f'http://skip.dev.hop.scimma.org/api/alerts/{test_alert["id"]}')
        self.assertEqual(generic_alert.score, '')

    def test_to_generic_alert_lvc_topic(self):
        """Test the to_generic_alert logic for lvc-counterpart alerts. Should result in the inclusion of score."""
        test_alert = self.alerts[0]
        test_alert['topic'] = 'lvc-counterpart'
        test_alert['message'] = {'rank': '3'}
        generic_alert = SCIMMABroker().to_generic_alert(test_alert)

        # NOTE: The string is hardcoded as a sanity check to ensure that the string is reviewed if it changes
        self.assertEqual(generic_alert.url, f'http://skip.dev.hop.scimma.org/api/alerts/{test_alert["id"]}')
        self.assertEqual(generic_alert.score, '3')

    def test_to_target_any_topic(self):
        """Test the SCIMMA-specific to_target logic."""
        test_alert = self.alerts[0]
        test_alert['topic'] = 'gcn'
        SCIMMABroker().to_target(test_alert)

        target = Target.objects.filter(name=test_alert['alert_identifier']).first()

        self.assertIsInstance(target, Target)
        self.assertEqual(target.galactic_lat, None)
        self.assertEqual(target.ra, self.alerts[0]['right_ascension'])
        self.assertEqual(target.dec, self.alerts[0]['declination'])

    def test_to_target_lvc_topic(self):
        """
        Test the to_target logic for lvc-counterpart alerts. Should result in the inclusion of galactic coordinates.
        """
        test_alert = self.alerts[0]
        test_alert['topic'] = 'lvc-counterpart'
        test_alert['message']['gal_coords'] = '336.99,-45.74 [deg] galactic lon,lat of the counterpart'
        SCIMMABroker().to_target(test_alert)

        target = Target.objects.filter(name=test_alert['alert_identifier']).first()

        self.assertIsInstance(target, Target)
        self.assertEqual(target.galactic_lat, -45.74)
        self.assertEqual(target.ra, self.alerts[0]['right_ascension'])
        self.assertEqual(target.dec, self.alerts[0]['declination'])

    def test_submit_upstream_alert_with_topic(self):
        """Test that successful submission uses the provided topic."""
        t = Target.objects.create(name='test name', ra=1, dec=2)
        with patch('tom_scimma.scimma.Stream.open', mock_open(read_data='data')) as mock_stream:
            SCIMMABroker().submit_upstream_alert(target=t, observation_record=None, topic='test')
            mock_stream.assert_called_once_with('kafka://test_url:9092/test', 'w')
            mock_stream().write.assert_called_with({'type': 'target', 'target_name': t.name, 'ra': t.ra, 'dec': t.dec})

    def test_submit_upstream_alert_no_topic(self):
        """Test that submission with no topic succeeds, using the default topic."""
        t = Target.objects.create(name='test name', ra=1, dec=2)
        with patch('tom_scimma.scimma.Stream.open', mock_open(read_data='data')) as mock_stream:
            SCIMMABroker().submit_upstream_alert(target=t, observation_record=None, topic=None)
            mock_stream.assert_called_once_with('kafka://test_url:9092/default', 'w')
            mock_stream().write.assert_called_with({'type': 'target', 'target_name': t.name, 'ra': t.ra, 'dec': t.dec})

    @override_settings(BROKERS={'SCIMMA': {'api_key': '', 'hopskotch_username': '', 'hopskotch_password': '',
                                           'hopskotch_url': 'test_url', 'default_hopskotch_topic': ''}
                                })
    @patch('tom_scimma.scimma.Stream')
    def test_submit_upstream_alert_no_default_topic(self, mock_stream):
        """Test that submission with no topic and no default topic fails."""
        with self.assertRaises(AlertSubmissionException):
            SCIMMABroker().submit_upstream_alert(target=None, observation_record=None)

    def test_submit_upstream_alert_failure(self):
        """Test that a KafkaException results in an AlertSubmissionException."""
        t = Target.objects.create(name='test name', ra=1, dec=2)
        with patch('tom_scimma.scimma.Stream.open', mock_open(read_data='data')) as mock_stream:
            mock_stream().write.side_effect = KafkaException()
            with self.assertRaises(AlertSubmissionException):
                SCIMMABroker().submit_upstream_alert(target=t, observation_record=None)
